import sys
import random
import numpy as np
from kalmanfilter import KalmanFilter 

class Car(object):

    def __init__(self, prev, speed, position, IPD, PS, tps):
        self.platoonPrev = prev
        self.cPos = position
        self.IPD  = IPD
        self.PS   = PS
        self.tps  = tps
        
        self.bcIPD = False
        
        # initialise calman filter
        if not prev == None:
            dt = 1/tps
            x = np.array([[0,0]]).T
            P = np.array([[10,0],[0,10]])

            F = np.array([[1,dt],[0,1]])
            Q = np.array([[1,0],[0,1]])
            H = np.array([[1,0]])
            R = np.array([[5]])

            self.KF = KalmanFilter(x, P, F, Q, H, R, tps)

            self.KF.predict()

        self.dIdx = 0
        self.distHistLength = 15
        self.ds = np.empty(self.distHistLength)
        
        RAND = 0.5
        self.SonarSystemUnc      = random.uniform(-RAND,RAND)       # in cm
        self.SonarStatisticalUnc = lambda : random.gauss(0,RAND) # in cm
        
        self.SpeedSystemUnc      = random.uniform(-RAND,RAND)        # in meter per seconds
        self.SpeedStatisticalUnc = lambda : random.gauss(0,RAND) # in meter per seconds
        
        self.EngineSystemUnc     = random.uniform(-RAND,RAND) # in meter per seconds
        
        self.setSpeed(speed)
        self.getDistance()
        self.v_v = self.cSpeed
        return None

    def getCIPD(self):
        return max(0.1*self.IPD,1)
    
    def getSpeed(self):
        if self.cSpeed <= 1e-8:
            return 0
        self.dIdx = (self.dIdx + 1) % self.distHistLength
        self.ds[self.dIdx] = self.SpeedSystemUnc + self.cSpeed + self.SpeedStatisticalUnc()
        return np.median(self.ds)
      
    def setSpeed(self,value):
        if value > 1e-8:
            self.cSpeed = value
            self.speed  = self.EngineSystemUnc + value
        else:
            self.cSpeed = 0
            self.speed = 0
      
    def getDistance(self):
        setLater = False
        try: 
            self.oldDist = self.dist
        except AttributeError:
            setLater = True
        
        if not self.platoonPrev == None:
            self.dist = self.SonarSystemUnc + (self.platoonPrev.cPos - self.cPos) + self.SonarStatisticalUnc()
        else:
            self.dist = sys.maxsize
            
        if setLater:
            self.oldDist = self.dist
        return self.dist

    def updatePos(self):
        self.cPos += self.cSpeed/float(self.tps)

    def updateSpeed(self):
        if not self.bcIPD:
            d = self.getDistance() # sensor data
            v = self.getSpeed()    # sesnor data
            IPD = self.IPD
            PS  = self.PS
            v_new = None

            if not self.platoonPrev ==  None:
                self.KF.update(np.array([[self.getDistance()]]))
                self.KF.predict()
                x = self.KF.getValue()
                d = x[0,0]
                self.v_v = v_v = max(v + x[1,0], 0)

                v_new = self.computeNewSpeed_2(d, v, v_v, IPD, PS)

                inertia = 0.8                                   # in [0,1] und gibt die Traegheit des Fahrzeuges an
                self.setSpeed(inertia * v + (1- inertia) * v_new)
                
                if self.getCIPD() > d:
                    print("raise")
                    raise UnderCIPDerror("under CIPD!")
                
            else: # this is for the LV
                c = 0.5
                self.v_v = 0
                self.setSpeed(c * v + (1-c) * PS)
                
                
    def computeNewSpeed_1(self, d, v, v_v, IPD, PS):
        if checkInbound(d, IPD, 0.01*IPD):
            if checkInbound (v, v_v, 0.01*v_v):
                if checkInbound(v, PS, 0.01*PS):
                    v_new = PS
                else:
                    if v > PS:
                        v_new = v_v - abs(PS-v_v)/4
                    else:
                        v_new = v_v + abs(PS-v_v)/4
            else:
                if v > v_v:
                    v_new = v_v - abs(PS-v_v)/4
                else:
                    v_new = v_v + abs(PS-v_v)/4
        else:
            v_new = (v_v * (d/IPD)**2)      #Der Exponent gibt an, wie schnell die Aenderung umgesetzt werden soll
        return v_new
    
    def computeNewSpeed_2(self, d, v, v_v, IPD, PS):
        # Der Exponent gibt an, wie schnell die Aenderung umgesetzt werden soll
        DIST_POW = 2
        SPEED_POW = 0.5
        if abs(v_v) > 1e-8:
          v_new = (v_v * (d/(IPD))**DIST_POW * (PS/v_v)**SPEED_POW) 
        else:
          v_new = (d - IPD)/IPD
        return v_new

    def computeNewSpeed_3(self, d, v, v_v, IPD, PS):
        # Der Exponent gibt an, wie schnell die Aenderung umgesetzt werden soll
        DIST_POW = 1.5; DIST_FAC = 1
        SPEED_POW = 0.5; SPEED_FAC = 1
        if abs(v_v) > 1e-8:
          v_new = (v_v * DIST_FAC*(d/(IPD))**DIST_POW * SPEED_FAC*(PS/v_v)**SPEED_POW) 
        else:
          v_new = (d - IPD)/IPD
        return v_new
